# -*- coding: utf-8 -*-

# project
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.ext import net
from ddtrace.contrib.flask_cache import get_traced_cache
from ddtrace.contrib.flask_cache.tracers import CACHE_BACKEND

# 3rd party
from flask import Flask

# testing
from tests.opentracer.utils import init_tracer
from ..config import REDIS_CONFIG, MEMCACHED_CONFIG
from ... import TracerTestCase, assert_is_measured, assert_dict_issuperset


class FlaskCacheTest(TracerTestCase):
    SERVICE = 'test-flask-cache'
    TEST_REDIS_PORT = REDIS_CONFIG['port']
    TEST_MEMCACHED_PORT = MEMCACHED_CONFIG['port']

    def setUp(self):
        super(FlaskCacheTest, self).setUp()

        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        self.cache = Cache(app, config={'CACHE_TYPE': 'simple'})

    def test_simple_cache_get(self):
        self.cache.get(u'á_complex_operation')
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'get')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        expected_meta = {
            'flask_cache.key': u'á_complex_operation',
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(span.meta, expected_meta)

    def test_simple_cache_set(self):
        self.cache.set(u'á_complex_operation', u'with_á_value\nin two lines')
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'set')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        expected_meta = {
            'flask_cache.key': u'á_complex_operation',
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(span.meta, expected_meta)

    def test_simple_cache_add(self):
        self.cache.add(u'á_complex_number', 50)
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'add')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        expected_meta = {
            'flask_cache.key': u'á_complex_number',
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(span.meta, expected_meta)

    def test_simple_cache_delete(self):
        self.cache.delete(u'á_complex_operation')
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'delete')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        expected_meta = {
            'flask_cache.key': u'á_complex_operation',
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(span.meta, expected_meta)

    def test_simple_cache_delete_many(self):
        self.cache.delete_many('complex_operation', 'another_complex_op')
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'delete_many')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        expected_meta = {
            'flask_cache.key': "['complex_operation', 'another_complex_op']",
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(span.meta, expected_meta)

    def test_simple_cache_clear(self):
        self.cache.clear()
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'clear')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        expected_meta = {
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(span.meta, expected_meta)

    def test_simple_cache_get_many(self):
        self.cache.get_many('first_complex_op', 'second_complex_op')
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'get_many')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        expected_meta = {
            'flask_cache.key': "['first_complex_op', 'second_complex_op']",
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(span.meta, expected_meta)

    def test_simple_cache_set_many(self):
        self.cache.set_many({
            'first_complex_op': 10,
            'second_complex_op': 20,
        })
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        assert_is_measured(span)
        self.assertEqual(span.service, self.SERVICE)
        self.assertEqual(span.resource, 'set_many')
        self.assertEqual(span.name, 'flask_cache.cmd')
        self.assertEqual(span.span_type, 'cache')
        self.assertEqual(span.error, 0)

        self.assertEqual(span.meta['flask_cache.backend'], 'simple')
        self.assertTrue('first_complex_op' in span.meta['flask_cache.key'])
        self.assertTrue('second_complex_op' in span.meta['flask_cache.key'])

    def test_default_span_tags(self):
        # test tags and attributes
        with self.cache._TracedCache__trace('flask_cache.cmd') as span:
            self.assertEqual(span.service, self.SERVICE)
            self.assertEqual(span.span_type, 'cache')
            self.assertEqual(span.meta[CACHE_BACKEND], 'simple')
            self.assertTrue(net.TARGET_HOST not in span.meta)
            self.assertTrue(net.TARGET_PORT not in span.meta)

    def test_default_span_tags_for_redis(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        config = {
            'CACHE_TYPE': 'redis',
            'CACHE_REDIS_PORT': self.TEST_REDIS_PORT,
        }
        cache = Cache(app, config=config)
        # test tags and attributes
        with cache._TracedCache__trace('flask_cache.cmd') as span:
            self.assertEqual(span.service, self.SERVICE)
            self.assertEqual(span.span_type, 'cache')
            self.assertEqual(span.meta[CACHE_BACKEND], 'redis')
            self.assertEqual(span.meta[net.TARGET_HOST], 'localhost')
            self.assertEqual(span.metrics[net.TARGET_PORT], self.TEST_REDIS_PORT)

    def test_default_span_tags_memcached(self):
        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        config = {
            'CACHE_TYPE': 'memcached',
            'CACHE_MEMCACHED_SERVERS': ['127.0.0.1:{}'.format(self.TEST_MEMCACHED_PORT)],
        }
        cache = Cache(app, config=config)
        # test tags and attributes
        with cache._TracedCache__trace('flask_cache.cmd') as span:
            self.assertEqual(span.service, self.SERVICE)
            self.assertEqual(span.span_type, 'cache')
            self.assertEqual(span.meta[CACHE_BACKEND], 'memcached')
            self.assertEqual(span.meta[net.TARGET_HOST], '127.0.0.1')
            self.assertEqual(span.metrics[net.TARGET_PORT], self.TEST_MEMCACHED_PORT)

    def test_simple_cache_get_ot(self):
        """OpenTracing version of test_simple_cache_get."""
        ot_tracer = init_tracer('my_svc', self.tracer)

        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer, service=self.SERVICE)
        app = Flask(__name__)
        cache = Cache(app, config={'CACHE_TYPE': 'simple'})

        with ot_tracer.start_active_span('ot_span'):
            cache.get(u'á_complex_operation')

        spans = self.get_spans()
        self.assertEqual(len(spans), 2)
        ot_span, dd_span = spans

        # confirm the parenting
        self.assertIsNone(ot_span.parent_id)
        self.assertEqual(dd_span.parent_id, ot_span.span_id)

        self.assertEqual(ot_span.resource, 'ot_span')
        self.assertEqual(ot_span.service, 'my_svc')

        assert_is_measured(dd_span)
        self.assertEqual(dd_span.service, self.SERVICE)
        self.assertEqual(dd_span.resource, 'get')
        self.assertEqual(dd_span.name, 'flask_cache.cmd')
        self.assertEqual(dd_span.span_type, 'cache')
        self.assertEqual(dd_span.error, 0)

        expected_meta = {
            'flask_cache.key': u'á_complex_operation',
            'flask_cache.backend': 'simple',
        }

        assert_dict_issuperset(dd_span.meta, expected_meta)

    def test_analytics_default(self):
        self.cache.get(u'á_complex_operation')
        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        self.assertIsNone(spans[0].get_metric(ANALYTICS_SAMPLE_RATE_KEY))

    def test_analytics_with_rate(self):
        with self.override_config(
            'flask_cache',
            dict(analytics_enabled=True, analytics_sample_rate=0.5)
        ):
            self.cache.get(u'á_complex_operation')

        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        self.assertEqual(spans[0].get_metric(ANALYTICS_SAMPLE_RATE_KEY), 0.5)

    def test_analytics_without_rate(self):
        with self.override_config(
            'flask_cache',
            dict(analytics_enabled=True)
        ):
            self.cache.get(u'á_complex_operation')

        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        self.assertEqual(spans[0].get_metric(ANALYTICS_SAMPLE_RATE_KEY), 1.0)


class TestFlaskCacheSettings(TracerTestCase):
    TEST_REDIS_PORT = REDIS_CONFIG['port']
    TEST_MEMCACHED_PORT = MEMCACHED_CONFIG['port']

    def setUp(self):
        super(TestFlaskCacheSettings, self).setUp()

        # create the TracedCache instance for a Flask app
        Cache = get_traced_cache(self.tracer)
        app = Flask(__name__)
        self.cache = Cache(app, config={'CACHE_TYPE': 'simple'})

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_SERVICE="mysvc"))
    def test_user_specified_service(self):
        """
        When a service name is specified by the user
            The flask-cache integration should use it as the service name
        """

        self.cache.get(u'á_complex_operation')
        spans = self.get_spans()

        for span in spans:
            span.service == "mysvc"



# -*- encoding: utf-8 -*-
import logging

from ddtrace import compat
from ddtrace.profiling import _attr
from ddtrace.profiling import _periodic
from ddtrace.profiling import _traceback
from ddtrace.profiling import exporter
from ddtrace.vendor import attr

LOG = logging.getLogger(__name__)


@attr.s
class Scheduler(_periodic.PeriodicService):
    """Schedule export of recorded data."""

    recorder = attr.ib()
    exporters = attr.ib()
    _interval = attr.ib(factory=_attr.from_env("DD_PROFILING_UPLOAD_INTERVAL", 60, float))
    _configured_interval = attr.ib(init=False)
    _last_export = attr.ib(init=False, default=None)

    def __attrs_post_init__(self):
        # Copy the value to use it later since we're going to adjust the real interval
        self._configured_interval = self.interval

    def start(self):
        """Start the scheduler."""
        LOG.debug("Starting scheduler")
        super(Scheduler, self).start()
        self._last_export = compat.time_ns()
        LOG.debug("Scheduler started")

    def flush(self):
        """Flush events from recorder to exporters."""
        LOG.debug("Flushing events")
        if self.exporters:
            events = self.recorder.reset()
            start = self._last_export
            self._last_export = compat.time_ns()
            for exp in self.exporters:
                try:
                    exp.export(events, start, self._last_export)
                except exporter.ExportError as e:
                    LOG.error("%s. Ignoring.", _traceback.format_exception(e))
                except Exception:
                    LOG.exception(
                        "Unexpected error while exporting events. "
                        "Please report this bug to https://github.com/DataDog/dd-trace-py/issues"
                    )

    def periodic(self):
        start_time = compat.monotonic()
        try:
            self.flush()
        finally:
            self.interval = max(0, self._configured_interval - (compat.monotonic() - start_time))

    on_shutdown = flush



from ...constants import SPAN_MEASURED_KEY
from ddtrace.pin import Pin
from ddtrace.settings import config
from ddtrace.utils.wrappers import unwrap as _u
from ddtrace.vendor.wrapt import wrap_function_wrapper as _w

from .. import trace_utils

DD_PATCH_ATTR = '_datadog_patch'

SERVICE_NAME = 'algoliasearch'
APP_NAME = 'algoliasearch'

try:
    import algoliasearch
    from algoliasearch.version import VERSION
    algoliasearch_version = tuple([int(i) for i in VERSION.split('.')])

    # Default configuration
    config._add('algoliasearch', dict(
        _default_service=SERVICE_NAME,
        collect_query_text=False
    ))
except ImportError:
    algoliasearch_version = (0, 0)


def patch():
    if algoliasearch_version == (0, 0):
        return

    if getattr(algoliasearch, DD_PATCH_ATTR, False):
        return

    setattr(algoliasearch, '_datadog_patch', True)

    pin = Pin(app=APP_NAME)

    if algoliasearch_version < (2, 0) and algoliasearch_version >= (1, 0):
        _w(algoliasearch.index, 'Index.search', _patched_search)
        pin.onto(algoliasearch.index.Index)
    elif algoliasearch_version >= (2, 0) and algoliasearch_version < (3, 0):
        from algoliasearch import search_index
        _w(algoliasearch, 'search_index.SearchIndex.search', _patched_search)
        pin.onto(search_index.SearchIndex)
    else:
        return


def unpatch():
    if algoliasearch_version == (0, 0):
        return

    if getattr(algoliasearch, DD_PATCH_ATTR, False):
        setattr(algoliasearch, DD_PATCH_ATTR, False)

    if algoliasearch_version < (2, 0) and algoliasearch_version >= (1, 0):
        _u(algoliasearch.index.Index, 'search')
    elif algoliasearch_version >= (2, 0) and algoliasearch_version < (3, 0):
        from algoliasearch import search_index
        _u(search_index.SearchIndex, 'search')
    else:
        return


# DEV: this maps serves the dual purpose of enumerating the algoliasearch.search() query_args that
# will be sent along as tags, as well as converting arguments names into tag names compliant with
# tag naming recommendations set out here: https://docs.datadoghq.com/tagging/
QUERY_ARGS_DD_TAG_MAP = {
    'page': 'page',
    'hitsPerPage': 'hits_per_page',
    'attributesToRetrieve': 'attributes_to_retrieve',
    'attributesToHighlight': 'attributes_to_highlight',
    'attributesToSnippet': 'attributes_to_snippet',
    'minWordSizefor1Typo': 'min_word_size_for_1_typo',
    'minWordSizefor2Typos': 'min_word_size_for_2_typos',
    'getRankingInfo': 'get_ranking_info',
    'aroundLatLng': 'around_lat_lng',
    'numericFilters': 'numeric_filters',
    'tagFilters': 'tag_filters',
    'queryType': 'query_type',
    'optionalWords': 'optional_words',
    'distinct': 'distinct'
}


def _patched_search(func, instance, wrapt_args, wrapt_kwargs):
    """
        wrapt_args is called the way it is to distinguish it from the 'args'
        argument to the algoliasearch.index.Index.search() method.
    """

    if algoliasearch_version < (2, 0) and algoliasearch_version >= (1, 0):
        function_query_arg_name = 'args'
    elif algoliasearch_version >= (2, 0) and algoliasearch_version < (3, 0):
        function_query_arg_name = 'request_options'
    else:
        return func(*wrapt_args, **wrapt_kwargs)

    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return func(*wrapt_args, **wrapt_kwargs)

    with pin.tracer.trace(
        'algoliasearch.search', service=trace_utils.ext_service(pin, config.algoliasearch)
    ) as span:
        span.set_tag(SPAN_MEASURED_KEY)

        if not span.sampled:
            return func(*wrapt_args, **wrapt_kwargs)

        if config.algoliasearch.collect_query_text:
            span.set_tag('query.text', wrapt_kwargs.get('query', wrapt_args[0]))

        query_args = wrapt_kwargs.get(function_query_arg_name, wrapt_args[1] if len(wrapt_args) > 1 else None)

        if query_args and isinstance(query_args, dict):
            for query_arg, tag_name in QUERY_ARGS_DD_TAG_MAP.items():
                value = query_args.get(query_arg)
                if value is not None:
                    span.set_tag('query.args.{}'.format(tag_name), value)

        # Result would look like this
        # {
        #   'hits': [
        #     {
        #       .... your search results ...
        #     }
        #   ],
        #   'processingTimeMS': 1,
        #   'nbHits': 1,
        #   'hitsPerPage': 20,
        #   'exhaustiveNbHits': true,
        #   'params': 'query=xxx',
        #   'nbPages': 1,
        #   'query': 'xxx',
        #   'page': 0
        # }
        result = func(*wrapt_args, **wrapt_kwargs)

        if isinstance(result, dict):
            if result.get('processingTimeMS', None) is not None:
                span.set_metric('processing_time_ms', int(result['processingTimeMS']))

            if result.get('nbHits', None) is not None:
                span.set_metric('number_of_hits', int(result['nbHits']))

        return result



# -*- encoding: utf-8 -*-
import sys
import threading

from ddtrace.profiling import _service
from ddtrace.profiling import _nogevent
from ddtrace.vendor import attr


PERIODIC_THREADS = {}


class PeriodicThread(threading.Thread):
    """Periodic thread.

    This class can be used to instantiate a worker thread that will run its `run_periodic` function every `interval`
    seconds.

    """

    def __init__(self, interval, target, name=None, on_shutdown=None):
        """Create a periodic thread.

        :param interval: The interval in seconds to wait between execution of the periodic function.
        :param target: The periodic function to execute every interval.
        :param name: The name of the thread.
        :param on_shutdown: The function to call when the thread shuts down.
        """
        super(PeriodicThread, self).__init__(name=name)
        self._target = target
        self._on_shutdown = on_shutdown
        self.interval = interval
        self.quit = threading.Event()
        self.daemon = True

    def stop(self):
        """Stop the thread."""
        # NOTE: make sure the thread is alive before using self.quit:
        # 1. self.quit is Lock-based
        # 2. if we're a child trying to stop a Thread,
        #    the Lock might have been locked in a parent process while forking so that'd block forever
        if self.is_alive():
            self.quit.set()

    def run(self):
        """Run the target function periodically."""
        PERIODIC_THREADS[self.ident] = self

        try:
            while not self.quit.wait(self.interval):
                self._target()
            if self._on_shutdown is not None:
                self._on_shutdown()
        finally:
            del PERIODIC_THREADS[self.ident]


class _GeventPeriodicThread(PeriodicThread):
    """Periodic thread.

    This class can be used to instantiate a worker thread that will run its `run_periodic` function every `interval`
    seconds.

    """

    # That's the value Python 2 uses in its `threading` module
    SLEEP_INTERVAL = 0.005

    def __init__(self, interval, target, name=None, on_shutdown=None):
        """Create a periodic thread.

        :param interval: The interval in seconds to wait between execution of the periodic function.
        :param target: The periodic function to execute every interval.
        :param name: The name of the thread.
        :param on_shutdown: The function to call when the thread shuts down.
        """
        super(_GeventPeriodicThread, self).__init__(interval, target, name, on_shutdown)
        self._tident = None

    @property
    def ident(self):
        return self._tident

    def start(self):
        """Start the thread."""
        self.quit = False
        self.has_quit = False
        self._tident = _nogevent.start_new_thread(self.run, tuple())
        if _nogevent.threading_get_native_id:
            self._native_id = _nogevent.threading_get_native_id()

    def join(self, timeout=None):
        # FIXME: handle the timeout argument
        while not self.has_quit:
            _nogevent.sleep(self.SLEEP_INTERVAL)

    def stop(self):
        """Stop the thread."""
        self.quit = True

    def run(self):
        """Run the target function periodically."""
        PERIODIC_THREADS[self._tident] = self

        try:
            while self.quit is False:
                self._target()
                slept = 0
                while self.quit is False and slept < self.interval:
                    _nogevent.sleep(self.SLEEP_INTERVAL)
                    slept += self.SLEEP_INTERVAL
            if self._on_shutdown is not None:
                self._on_shutdown()
        except Exception:
            # Exceptions might happen during interpreter shutdown.
            # We're mimicking what `threading.Thread` does in daemon mode, we ignore them.
            # See `threading.Thread._bootstrap` for details.
            if sys is not None:
                raise
        finally:
            try:
                self.has_quit = True
                del PERIODIC_THREADS[self._tident]
            except Exception:
                # Exceptions might happen during interpreter shutdown.
                # We're mimicking what `threading.Thread` does in daemon mode, we ignore them.
                # See `threading.Thread._bootstrap` for details.
                if sys is not None:
                    raise


def PeriodicRealThread(*args, **kwargs):
    """Create a PeriodicRealThread based on the underlying thread implementation (native, gevent, etc).

    This is exactly like PeriodicThread, except that it runs on a *real* OS thread. Be aware that this might be tricky
    in e.g. the gevent case, where Lock object must not be shared with the MainThread (otherwise it'd dead lock).

    """
    if _nogevent.is_module_patched("threading"):
        return _GeventPeriodicThread(*args, **kwargs)
    return PeriodicThread(*args, **kwargs)


@attr.s
class PeriodicService(_service.Service):
    """A service that runs periodically."""

    _interval = attr.ib()
    _worker = attr.ib(default=None, init=False, repr=False)

    _real_thread = False
    "Class variable to override if the service should run in a real OS thread."

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, value):
        self._interval = value
        # Update the interval of the PeriodicThread based on ours
        if self._worker:
            self._worker.interval = value

    def start(self):
        """Start the periodic service."""
        super(PeriodicService, self).start()
        periodic_thread_class = PeriodicRealThread if self._real_thread else PeriodicThread
        self._worker = periodic_thread_class(
            self.interval,
            target=self.periodic,
            name="%s:%s" % (self.__class__.__module__, self.__class__.__name__),
            on_shutdown=self.on_shutdown,
        )
        self._worker.start()

    def join(self, timeout=None):
        if self._worker:
            self._worker.join(timeout)

    def stop(self):
        """Stop the periodic collector."""
        if self._worker:
            self._worker.stop()
        super(PeriodicService, self).stop()

    @staticmethod
    def on_shutdown():
        pass

    @staticmethod
    def periodic():
        pass



import aiobotocore
from botocore.errorfactory import ClientError

from ddtrace.contrib.aiobotocore.patch import patch, unpatch
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.compat import stringify

from .utils import aiobotocore_client
from ..asyncio.utils import AsyncioTestCase, mark_asyncio
from tests.tracer.test_tracer import get_dummy_tracer
from ... import assert_is_measured, assert_span_http_status_code


class AIOBotocoreTest(AsyncioTestCase):
    """Botocore integration testsuite"""

    def setUp(self):
        super(AIOBotocoreTest, self).setUp()
        patch()
        self.tracer = get_dummy_tracer()

    def tearDown(self):
        super(AIOBotocoreTest, self).tearDown()
        unpatch()
        self.tracer = None

    @mark_asyncio
    def test_traced_client(self):
        with aiobotocore_client("ec2", self.tracer) as ec2:
            yield from ec2.describe_instances()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)
        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.get_tag("aws.agent"), "aiobotocore")
        self.assertEqual(span.get_tag("aws.region"), "us-west-2")
        self.assertEqual(span.get_tag("aws.operation"), "DescribeInstances")
        assert_span_http_status_code(span, 200)
        self.assertEqual(span.get_metric("retry_attempts"), 0)
        self.assertEqual(span.service, "aws.ec2")
        self.assertEqual(span.resource, "ec2.describeinstances")
        self.assertEqual(span.name, "ec2.command")
        self.assertEqual(span.span_type, "http")
        self.assertIsNone(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY))

    @mark_asyncio
    def test_traced_client_analytics(self):
        with self.override_config("aiobotocore", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
            with aiobotocore_client("ec2", self.tracer) as ec2:
                yield from ec2.describe_instances()

        spans = self.get_spans()
        assert spans
        span = spans[0]
        self.assertEqual(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 0.5)

    @mark_asyncio
    def test_s3_client(self):
        with aiobotocore_client("s3", self.tracer) as s3:
            yield from s3.list_buckets()
            yield from s3.list_buckets()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 2)
        self.assertEqual(len(traces[0]), 1)
        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.get_tag("aws.operation"), "ListBuckets")
        assert_span_http_status_code(span, 200)
        self.assertEqual(span.service, "aws.s3")
        self.assertEqual(span.resource, "s3.listbuckets")
        self.assertEqual(span.name, "s3.command")

    @mark_asyncio
    def test_s3_put(self):
        params = dict(Key="foo", Bucket="mybucket", Body=b"bar")

        with aiobotocore_client("s3", self.tracer) as s3:
            yield from s3.create_bucket(Bucket="mybucket")
            yield from s3.put_object(**params)

        spans = [trace[0] for trace in self.tracer.writer.pop_traces()]
        assert spans
        self.assertEqual(len(spans), 2)
        self.assertEqual(spans[0].get_tag("aws.operation"), "CreateBucket")

        assert_is_measured(spans[0])
        assert_span_http_status_code(spans[0], 200)
        self.assertEqual(spans[0].service, "aws.s3")
        self.assertEqual(spans[0].resource, "s3.createbucket")

        assert_is_measured(spans[1])
        self.assertEqual(spans[1].get_tag("aws.operation"), "PutObject")
        self.assertEqual(spans[1].resource, "s3.putobject")
        self.assertEqual(spans[1].get_tag("params.Key"), stringify(params["Key"]))
        self.assertEqual(spans[1].get_tag("params.Bucket"), stringify(params["Bucket"]))
        self.assertIsNone(spans[1].get_tag("params.Body"))

    @mark_asyncio
    def test_s3_client_error(self):
        with aiobotocore_client("s3", self.tracer) as s3:
            with self.assertRaises(ClientError):
                # FIXME: add proper clean-up to tearDown
                yield from s3.list_objects(Bucket="doesnotexist")

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)
        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.resource, "s3.listobjects")
        self.assertEqual(span.error, 1)
        self.assertTrue("NoSuchBucket" in span.get_tag("error.msg"))

    @mark_asyncio
    def test_s3_client_read(self):
        with aiobotocore_client("s3", self.tracer) as s3:
            # prepare S3 and flush traces if any
            yield from s3.create_bucket(Bucket="tracing")
            yield from s3.put_object(Bucket="tracing", Key="apm", Body=b"")
            self.tracer.writer.pop_traces()
            # calls under test
            response = yield from s3.get_object(Bucket="tracing", Key="apm")
            yield from response["Body"].read()

        traces = self.tracer.writer.pop_traces()
        version = aiobotocore.__version__.split(".")
        pre_08 = int(version[0]) == 0 and int(version[1]) < 8
        if pre_08:
            self.assertEqual(len(traces), 2)
            self.assertEqual(len(traces[1]), 1)
        else:
            self.assertEqual(len(traces), 1)

        self.assertEqual(len(traces[0]), 1)

        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.get_tag("aws.operation"), "GetObject")
        assert_span_http_status_code(span, 200)
        self.assertEqual(span.service, "aws.s3")
        self.assertEqual(span.resource, "s3.getobject")

        if pre_08:
            read_span = traces[1][0]
            self.assertEqual(read_span.get_tag("aws.operation"), "GetObject")
            assert_span_http_status_code(read_span, 200)
            self.assertEqual(read_span.service, "aws.s3")
            self.assertEqual(read_span.resource, "s3.getobject")
            self.assertEqual(read_span.name, "s3.command.read")
            # enforce parenting
            self.assertEqual(read_span.parent_id, span.span_id)
            self.assertEqual(read_span.trace_id, span.trace_id)

    @mark_asyncio
    def test_sqs_client(self):
        with aiobotocore_client("sqs", self.tracer) as sqs:
            yield from sqs.list_queues()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)

        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.get_tag("aws.region"), "us-west-2")
        self.assertEqual(span.get_tag("aws.operation"), "ListQueues")
        assert_span_http_status_code(span, 200)
        self.assertEqual(span.service, "aws.sqs")
        self.assertEqual(span.resource, "sqs.listqueues")

    @mark_asyncio
    def test_kinesis_client(self):
        with aiobotocore_client("kinesis", self.tracer) as kinesis:
            yield from kinesis.list_streams()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)

        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.get_tag("aws.region"), "us-west-2")
        self.assertEqual(span.get_tag("aws.operation"), "ListStreams")
        assert_span_http_status_code(span, 200)
        self.assertEqual(span.service, "aws.kinesis")
        self.assertEqual(span.resource, "kinesis.liststreams")

    @mark_asyncio
    def test_lambda_client(self):
        with aiobotocore_client("lambda", self.tracer) as lambda_client:
            # https://github.com/spulec/moto/issues/906
            yield from lambda_client.list_functions(MaxItems=5)

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)

        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.get_tag("aws.region"), "us-west-2")
        self.assertEqual(span.get_tag("aws.operation"), "ListFunctions")
        assert_span_http_status_code(span, 200)
        self.assertEqual(span.service, "aws.lambda")
        self.assertEqual(span.resource, "lambda.listfunctions")

    @mark_asyncio
    def test_kms_client(self):
        with aiobotocore_client("kms", self.tracer) as kms:
            yield from kms.list_keys(Limit=21)

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)

        span = traces[0][0]

        assert_is_measured(span)
        self.assertEqual(span.get_tag("aws.region"), "us-west-2")
        self.assertEqual(span.get_tag("aws.operation"), "ListKeys")
        assert_span_http_status_code(span, 200)
        self.assertEqual(span.service, "aws.kms")
        self.assertEqual(span.resource, "kms.listkeys")
        # checking for protection on STS against security leak
        self.assertEqual(span.get_tag("params"), None)

    @mark_asyncio
    def test_unpatch(self):
        unpatch()
        with aiobotocore_client("kinesis", self.tracer) as kinesis:
            yield from kinesis.list_streams()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 0)

    @mark_asyncio
    def test_double_patch(self):
        patch()
        with aiobotocore_client("sqs", self.tracer) as sqs:
            yield from sqs.list_queues()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)

    @mark_asyncio
    def test_opentraced_client(self):
        from tests.opentracer.utils import init_tracer

        ot_tracer = init_tracer("my_svc", self.tracer)

        with ot_tracer.start_active_span("ot_outer_span"):
            with aiobotocore_client("ec2", self.tracer) as ec2:
                yield from ec2.describe_instances()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 2)
        ot_span = traces[0][0]
        dd_span = traces[0][1]

        self.assertEqual(ot_span.resource, "ot_outer_span")
        self.assertEqual(ot_span.service, "my_svc")

        # confirm the parenting
        self.assertEqual(ot_span.parent_id, None)
        self.assertEqual(dd_span.parent_id, ot_span.span_id)

        assert_is_measured(dd_span)
        self.assertEqual(dd_span.get_tag("aws.agent"), "aiobotocore")
        self.assertEqual(dd_span.get_tag("aws.region"), "us-west-2")
        self.assertEqual(dd_span.get_tag("aws.operation"), "DescribeInstances")
        assert_span_http_status_code(dd_span, 200)
        self.assertEqual(dd_span.get_metric("retry_attempts"), 0)
        self.assertEqual(dd_span.service, "aws.ec2")
        self.assertEqual(dd_span.resource, "ec2.describeinstances")
        self.assertEqual(dd_span.name, "ec2.command")

    @mark_asyncio
    def test_opentraced_s3_client(self):
        from tests.opentracer.utils import init_tracer

        ot_tracer = init_tracer("my_svc", self.tracer)

        with ot_tracer.start_active_span("ot_outer_span"):
            with aiobotocore_client("s3", self.tracer) as s3:
                yield from s3.list_buckets()
                with ot_tracer.start_active_span("ot_inner_span1"):
                    yield from s3.list_buckets()
                with ot_tracer.start_active_span("ot_inner_span2"):
                    pass

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 5)
        ot_outer_span = traces[0][0]
        dd_span = traces[0][1]
        ot_inner_span = traces[0][2]
        dd_span2 = traces[0][3]
        ot_inner_span2 = traces[0][4]

        self.assertEqual(ot_outer_span.resource, "ot_outer_span")
        self.assertEqual(ot_inner_span.resource, "ot_inner_span1")
        self.assertEqual(ot_inner_span2.resource, "ot_inner_span2")

        # confirm the parenting
        self.assertEqual(ot_outer_span.parent_id, None)
        self.assertEqual(dd_span.parent_id, ot_outer_span.span_id)
        self.assertEqual(ot_inner_span.parent_id, ot_outer_span.span_id)
        self.assertEqual(dd_span2.parent_id, ot_inner_span.span_id)
        self.assertEqual(ot_inner_span2.parent_id, ot_outer_span.span_id)

        assert_is_measured(dd_span)
        self.assertEqual(dd_span.get_tag("aws.operation"), "ListBuckets")
        assert_span_http_status_code(dd_span, 200)
        self.assertEqual(dd_span.service, "aws.s3")
        self.assertEqual(dd_span.resource, "s3.listbuckets")
        self.assertEqual(dd_span.name, "s3.command")

        self.assertEqual(dd_span2.get_tag("aws.operation"), "ListBuckets")
        assert_span_http_status_code(dd_span2, 200)
        self.assertEqual(dd_span2.service, "aws.s3")
        self.assertEqual(dd_span2.resource, "s3.listbuckets")
        self.assertEqual(dd_span2.name, "s3.command")

    @mark_asyncio
    @AsyncioTestCase.run_in_subprocess(env_overrides=dict(DD_SERVICE="mysvc"))
    def test_user_specified_service(self):
        """
        When a service name is specified by the user
            The aiobotocore integration should use it as the service name
        """
        with aiobotocore_client("ec2", self.tracer) as ec2:
            yield from ec2.describe_instances()

        traces = self.tracer.writer.pop_traces()
        self.assertEqual(len(traces), 1)
        self.assertEqual(len(traces[0]), 1)
        span = traces[0][0]

        assert span.service == "mysvc"



# stdlib
import itertools
import os
import sys
import threading
import time

from .. import api
from .. import compat
from .. import _worker
from ..internal.logger import get_logger
from ..sampler import BasePrioritySampler
from ..settings import config
from ..encoding import JSONEncoderV2
from ..payload import PayloadFull
from . import _queue

log = get_logger(__name__)


DEFAULT_TIMEOUT = 5
LOG_ERR_INTERVAL = 60


def _apply_filters(filters, traces):
    """
    Here we make each trace go through the filters configured in the
    tracer. There is no need for a lock since the traces are owned by the
    AgentWriter at that point.
    """
    if filters is not None:
        filtered_traces = []
        for trace in traces:
            for filtr in filters:
                trace = filtr.process_trace(trace)
                if trace is None:
                    break
            if trace is not None:
                filtered_traces.append(trace)
        return filtered_traces
    return traces


class LogWriter:
    def __init__(self, out=sys.stdout, filters=None, sampler=None, priority_sampler=None):
        self._filters = filters
        self._sampler = sampler
        self._priority_sampler = priority_sampler
        self.encoder = JSONEncoderV2()
        self.out = out

    def recreate(self):
        """Create a new instance of :class:`LogWriter` using the same settings from this instance

        :rtype: :class:`LogWriter`
        :returns: A new :class:`LogWriter` instance
        """
        writer = self.__class__(
            out=self.out, filters=self._filters, sampler=self._sampler, priority_sampler=self._priority_sampler
        )
        return writer

    def write(self, spans=None, services=None):
        # We immediately flush all spans
        if not spans:
            return

        # Before logging the traces, make them go through the
        # filters
        try:
            traces = _apply_filters(self._filters, [spans])
        except Exception:
            log.error("error while filtering traces", exc_info=True)
            return
        if len(traces) == 0:
            return
        encoded = self.encoder.encode_traces(traces)
        self.out.write(encoded + "\n")
        self.out.flush()


class AgentWriter(_worker.PeriodicWorkerThread):

    QUEUE_PROCESSING_INTERVAL = 1
    QUEUE_MAX_TRACES_DEFAULT = 1000

    def __init__(
        self,
        hostname="localhost",
        port=8126,
        uds_path=None,
        https=False,
        shutdown_timeout=DEFAULT_TIMEOUT,
        filters=None,
        sampler=None,
        priority_sampler=None,
        dogstatsd=None,
    ):
        super(AgentWriter, self).__init__(
            interval=self.QUEUE_PROCESSING_INTERVAL, exit_timeout=shutdown_timeout, name=self.__class__.__name__
        )
        # DEV: provide a _temporary_ solution to allow users to specify a custom max
        maxsize = int(os.getenv("DD_TRACE_MAX_TPS", self.QUEUE_MAX_TRACES_DEFAULT))
        self._trace_queue = _queue.TraceQueue(maxsize=maxsize)
        self._filters = filters
        self._sampler = sampler
        self._priority_sampler = priority_sampler
        self._last_error_ts = 0
        self.dogstatsd = dogstatsd
        self.api = api.API(
            hostname, port, uds_path=uds_path, https=https, priority_sampling=priority_sampler is not None
        )
        if hasattr(time, "thread_time"):
            self._last_thread_time = time.thread_time()
        self._started = False
        self._started_lock = threading.Lock()

    def recreate(self):
        """Create a new instance of :class:`AgentWriter` using the same settings from this instance

        :rtype: :class:`AgentWriter`
        :returns: A new :class:`AgentWriter` instance
        """
        writer = self.__class__(
            hostname=self.api.hostname,
            port=self.api.port,
            uds_path=self.api.uds_path,
            https=self.api.https,
            shutdown_timeout=self.exit_timeout,
            filters=self._filters,
            priority_sampler=self._priority_sampler,
            dogstatsd=self.dogstatsd,
        )
        return writer

    @property
    def _send_stats(self):
        """Determine if we're sending stats or not."""
        return bool(config.health_metrics_enabled and self.dogstatsd)

    def write(self, spans=None, services=None):
        # Start the AgentWriter on first write.
        # Starting it earlier might be an issue with gevent, see:
        # https://github.com/DataDog/dd-trace-py/issues/1192
        if self._started is False:
            with self._started_lock:
                if self._started is False:
                    self.start()
                    self._started = True
        if spans:
            self._trace_queue.put(spans)

    def flush_queue(self):
        traces = self._trace_queue.get()

        if not traces:
            return

        if self._send_stats:
            traces_queue_length = len(traces)
            traces_queue_spans = sum(map(len, traces))

        # Before sending the traces, make them go through the
        # filters
        try:
            traces = _apply_filters(self._filters, traces)
        except Exception:
            log.error("error while filtering traces", exc_info=True)
            return

        if self._send_stats:
            traces_filtered = len(traces) - traces_queue_length

        # If we have data, let's try to send it.
        traces_responses = self.api.send_traces(traces)
        for response in traces_responses:
            if not isinstance(response, PayloadFull):
                if isinstance(response, Exception) or response.status >= 400:
                    self._log_error_status(response)
                elif self._priority_sampler or isinstance(self._sampler, BasePrioritySampler):
                    result_traces_json = response.get_json()
                    if result_traces_json and "rate_by_service" in result_traces_json:
                        if self._priority_sampler:
                            self._priority_sampler.update_rate_by_service_sample_rates(
                                result_traces_json["rate_by_service"],
                            )
                        if isinstance(self._sampler, BasePrioritySampler):
                            self._sampler.update_rate_by_service_sample_rates(
                                result_traces_json["rate_by_service"],
                            )

        # Dump statistics
        # NOTE: Do not use the buffering of dogstatsd as it's not thread-safe
        # https://github.com/DataDog/datadogpy/issues/439
        if self._send_stats:
            # Statistics about the queue length, size and number of spans
            self.dogstatsd.increment("datadog.tracer.flushes")
            self._histogram_with_total("datadog.tracer.flush.traces", traces_queue_length)
            self._histogram_with_total("datadog.tracer.flush.spans", traces_queue_spans)

            # Statistics about the filtering
            self._histogram_with_total("datadog.tracer.flush.traces_filtered", traces_filtered)

            # Statistics about API
            self._histogram_with_total("datadog.tracer.api.requests", len(traces_responses))

            self._histogram_with_total(
                "datadog.tracer.api.errors",
                len(list(t for t in traces_responses if isinstance(t, Exception) and not isinstance(t, PayloadFull))),
            )

            self._histogram_with_total(
                "datadog.tracer.api.traces_payloadfull",
                len(list(t for t in traces_responses if isinstance(t, PayloadFull))),
            )

            for status, grouped_responses in itertools.groupby(
                sorted((t for t in traces_responses if not isinstance(t, Exception)), key=lambda r: r.status),
                key=lambda r: r.status,
            ):
                self._histogram_with_total(
                    "datadog.tracer.api.responses", len(list(grouped_responses)), tags=["status:%d" % status]
                )

            # Statistics about the writer thread
            if hasattr(time, "thread_time"):
                new_thread_time = time.thread_time()
                diff = new_thread_time - self._last_thread_time
                self._last_thread_time = new_thread_time
                self.dogstatsd.histogram("datadog.tracer.writer.cpu_time", diff)

    def _histogram_with_total(self, name, value, tags=None):
        """Helper to add metric as a histogram and with a `.total` counter"""
        self.dogstatsd.histogram(name, value, tags=tags)
        self.dogstatsd.increment("%s.total" % (name,), value, tags=tags)

    def run_periodic(self):
        if self._send_stats:
            self.dogstatsd.gauge("datadog.tracer.heartbeat", 1)

        try:
            self.flush_queue()
        finally:
            if not self._send_stats:
                return

            # Statistics about the rate at which spans are inserted in the queue
            dropped, enqueued, enqueued_lengths = self._trace_queue.pop_stats()
            self.dogstatsd.gauge("datadog.tracer.queue.max_length", self._trace_queue.maxsize)
            self.dogstatsd.increment("datadog.tracer.queue.dropped.traces", dropped)
            self.dogstatsd.increment("datadog.tracer.queue.enqueued.traces", enqueued)
            self.dogstatsd.increment("datadog.tracer.queue.enqueued.spans", enqueued_lengths)

    def on_shutdown(self):
        try:
            self.run_periodic()
        finally:
            if not self._send_stats:
                return

            self.dogstatsd.increment("datadog.tracer.shutdown")

    def _log_error_status(self, response):
        log_level = log.debug
        now = compat.monotonic()
        if now > self._last_error_ts + LOG_ERR_INTERVAL:
            log_level = log.error
            self._last_error_ts = now
        prefix = "Failed to send traces to Datadog Agent at %s: "
        if isinstance(response, api.Response):
            log_level(
                prefix + "HTTP error status %s, reason %s, message %s",
                self.api,
                response.status,
                response.reason,
                response.msg,
            )
        else:
            log_level(
                prefix + "%s",
                self.api,
                response,
            )



"""
Datadog trace code for flask_cache
"""

# stdlib
import logging

# project
from .utils import _extract_conn_tags, _resource_from_cache_prefix
from ...constants import ANALYTICS_SAMPLE_RATE_KEY, SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...settings import config

# 3rd party
from flask.ext.cache import Cache


log = logging.Logger(__name__)

DEFAULT_SERVICE = config.service or "flask-cache"

# standard tags
COMMAND_KEY = 'flask_cache.key'
CACHE_BACKEND = 'flask_cache.backend'
CONTACT_POINTS = 'flask_cache.contact_points'


def get_traced_cache(ddtracer, service=DEFAULT_SERVICE, meta=None):
    """
    Return a traced Cache object that behaves exactly as the ``flask.ext.cache.Cache class``
    """

    class TracedCache(Cache):
        """
        Traced cache backend that monitors any operations done by flask_cache. Observed actions are:
        * get, set, add, delete, clear
        * all ``many_`` operations
        """
        _datadog_tracer = ddtracer
        _datadog_service = service
        _datadog_meta = meta

        def __trace(self, cmd):
            """
            Start a tracing with default attributes and tags
            """
            # create a new span
            s = self._datadog_tracer.trace(cmd, span_type=SpanTypes.CACHE, service=self._datadog_service)
            s.set_tag(SPAN_MEASURED_KEY)
            # set span tags
            s.set_tag(CACHE_BACKEND, self.config.get('CACHE_TYPE'))
            s.set_tags(self._datadog_meta)
            # set analytics sample rate
            s.set_tag(
                ANALYTICS_SAMPLE_RATE_KEY,
                config.flask_cache.get_analytics_sample_rate()
            )
            # add connection meta if there is one
            if getattr(self.cache, '_client', None):
                try:
                    s.set_tags(_extract_conn_tags(self.cache._client))
                except Exception:
                    log.debug('error parsing connection tags', exc_info=True)

            return s

        def get(self, *args, **kwargs):
            """
            Track ``get`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('GET', self.config)
                if len(args) > 0:
                    span.set_tag(COMMAND_KEY, args[0])
                return super(TracedCache, self).get(*args, **kwargs)

        def set(self, *args, **kwargs):
            """
            Track ``set`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('SET', self.config)
                if len(args) > 0:
                    span.set_tag(COMMAND_KEY, args[0])
                return super(TracedCache, self).set(*args, **kwargs)

        def add(self, *args, **kwargs):
            """
            Track ``add`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('ADD', self.config)
                if len(args) > 0:
                    span.set_tag(COMMAND_KEY, args[0])
                return super(TracedCache, self).add(*args, **kwargs)

        def delete(self, *args, **kwargs):
            """
            Track ``delete`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('DELETE', self.config)
                if len(args) > 0:
                    span.set_tag(COMMAND_KEY, args[0])
                return super(TracedCache, self).delete(*args, **kwargs)

        def delete_many(self, *args, **kwargs):
            """
            Track ``delete_many`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('DELETE_MANY', self.config)
                span.set_tag(COMMAND_KEY, list(args))
                return super(TracedCache, self).delete_many(*args, **kwargs)

        def clear(self, *args, **kwargs):
            """
            Track ``clear`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('CLEAR', self.config)
                return super(TracedCache, self).clear(*args, **kwargs)

        def get_many(self, *args, **kwargs):
            """
            Track ``get_many`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('GET_MANY', self.config)
                span.set_tag(COMMAND_KEY, list(args))
                return super(TracedCache, self).get_many(*args, **kwargs)

        def set_many(self, *args, **kwargs):
            """
            Track ``set_many`` operation
            """
            with self.__trace('flask_cache.cmd') as span:
                span.resource = _resource_from_cache_prefix('SET_MANY', self.config)
                if len(args) > 0:
                    span.set_tag(COMMAND_KEY, list(args[0].keys()))
                return super(TracedCache, self).set_many(*args, **kwargs)

    return TracedCache



import abc
from ddtrace.vendor import six

from ddtrace.compat import contextvars
from .logger import get_logger
from ..context import Context

log = get_logger(__name__)

_DD_CONTEXTVAR = contextvars.ContextVar("datadog_contextvar", default=None)


class BaseContextManager(six.with_metaclass(abc.ABCMeta)):
    def __init__(self, reset=True):
        if reset:
            self.reset()

    @abc.abstractmethod
    def _has_active_context(self):
        pass

    @abc.abstractmethod
    def set(self, ctx):
        pass

    @abc.abstractmethod
    def get(self):
        pass

    def reset(self):
        pass


class ContextVarContextManager(BaseContextManager):
    def _has_active_context(self):
        ctx = _DD_CONTEXTVAR.get()
        return ctx is not None

    def set(self, ctx):
        _DD_CONTEXTVAR.set(ctx)

    def get(self):
        ctx = _DD_CONTEXTVAR.get()
        if not ctx:
            ctx = Context()
            self.set(ctx)

        return ctx

    def reset(self):
        _DD_CONTEXTVAR.set(None)


DefaultContextManager = ContextVarContextManager
